// EXPERT.CPP: a tiny forward-chaining expert system lookup engine
// (c) 1990 C Gazette. See notice in NOE.CPP for terms.

#include "node.hpp"
#include "sysio.hpp"  // system-dependent I/O functions
#include <ctype.h>
#include <stdio.h>

const bsize = 150;  // size of input line buffer
// tabs, commas and newlines delimit attributes:
const char * attrib_delimiter = ",\t\n";

node names("names"), attributes("attributes");

main(int argc, char * argv[]) 
{
  if(argc < 2) 
  {
    fprintf(stderr, "usage: expert database_file\n");
    exit(0);
  }
  
  FILE * infile = fopen(argv[1], "r");

  if(infile == NULL) 
  {
    fprintf(stderr, "cannot open %s\n", argv[1]);
    exit(1);
  }

  char buf[bsize];
  while (fgets(buf, bsize, infile))  // get each line in the database
  {
    *strchr(buf, '\n') = '\0';   // remove terminating newline
    if(buf[0] == '#') 
       continue;  // ignore comments (lines starting with '#')
    if(buf[0] == ':')            // start a new name
    {
        names.add(new node(&buf[1]));  // assume no duplicate names
        continue;                      // go get next line
    }
    // Not a comment or a new name, so treat line as attributes for the 
    // current name; parse using the ANSI library function strtok():
    char * token = strtok(buf, attrib_delimiter);  // first token
    if(token) do 
    {
      while(*token == ' ' || *token == '\t') 
             token++;  // move past whitespace
      if(!*token) 
             continue;  // empty space at end of line
      node * attr = attributes.contains(token);  // see if it's already there
      if(!attr) 
             attr = attributes.add(new node(token));  // if not, add one
      names.top()->add(attr);  // the forward link
      attr->add(names.top());  // the reverse link
    } 
      while((token = strtok(0, attrib_delimiter)) != NULL); // rest of tokens
  }
  char * menu =
  "-------------------------\n"
  "Expert system search menu"
  "\ns: start a new search\n"
  "n: new attribute for current search\n"
  "p: go back to previous level of search sequence (remove current level)\n"
  "b: look back one level in search sequence\n"
  "e: return to the end of the search sequence\n"
  "v: view attributes of current search results\n"
  "q: quit";

  node * search = new node("search");

  while(1) 
  {
    puts(menu); 
    printf("selection->");
    switch (tolower(get_key())) 
    {
      case 's'  : {   // start a new search
                    delete search;
                    search = new node("search");
                    clearscreen();
                    node * match = selection_menu( &attributes,
                      "choose an attribute to search for:");
                    search->add(new node(match->identity));
                    // copy the match list to the node we just created:
                    search->top()->copy(match);
                  }
                  break;
      case 'n'  : { // narrow the search (add a new attribute)
                    clearscreen();
                    node * match = selection_menu( &attributes,
                      "choose an attribute to narrow the search:");
                    search->add(new node(match->identity));
                    search->top()->copy(match);
                    node * previous = search->next();
                    // create the intersection between the two sets:
                    search->top()->intersect(previous);
                  }
                  break;
      case 'p'  : search->conditional_pop();  // always leave the top node
                  break;
      case 'b'  : search->next_valid();
                  break;
      case 'e'  : search->top();
                  break;
      case 'v'  : { // view individual items in search results
                    clearscreen();
                    node * select =
                    selection_menu(search->cursor->data, "item to view:");
                    clearscreen();
                    select->print();
                    puts("any key to continue"); get_key();
                  }
                  break;
      case 'q'  : exit(0);
      default   : continue;
    }
    clearscreen();
    puts("current search results:");
    search->cursor->data->print();
  }
}